"""Find nodata value of datasets."""

import click
import numpy as np
import rasterio
from scipy.stats import mode

from rio_alpha.utils import (
    _convert_rgb,
    _compute_continuous,
    _debug_mode,
    _search_image_edge,
    _evaluate_count,
)


def discover_ndv(rgb_orig, debug, verbose):
    """Returns nodata value by calculating mode of RGB array

    Parameters
    ----------
    rgb_orig: ndarray
              array of input pixels of shape (rows, cols, depth)
    debug: Boolean
           Enables matplotlib & printing of figures
    verbose: Boolean
             Prints extra information, like competing candidate values


    Returns
    -------
    list of nodata value candidates or empty string if none found

    """
    rgb_mod, rgb_mod_flat = _convert_rgb(rgb_orig)
    # Full image mode bincount
    mode_vals = mode(rgb_mod_flat)
    candidate_original = [int((mode_vals[0])[0, i]) for i in range(3)]

    # Find continuous values in RGB array
    candidate_continuous, arr = _compute_continuous(rgb_mod, 1)

    # If debug mode, print histograms & be verbose
    if debug:
        click.echo("Original image ndv candidate: %s" % (str(candidate_original)))
        click.echo("Filtered image ndv candidate: %s" % (str(candidate_continuous)))
        outplot = "/tmp/hist_plot.png"
        _debug_mode(rgb_mod_flat, arr, outplot)

    # Compare ndv candidates from full & squished image
    candidate_list = [
        i for i, j in zip(candidate_original, candidate_continuous) if i == j
    ]

    # If candidates from original & filtered images match exactly,
    # print value & exit
    if len(candidate_list) == 3:
        return candidate_list

    # If candidates do not match exactly, continue vetting process
    # by searching image edge for frequency of each candidate
    elif len(candidate_list) < 3:
        if verbose:
            click.echo(
                "Competing ndv candidates...searching "
                "image collar for value frequency. "
                "Candidate list: %s" % str(candidate_list)
            )

        count_img_edge_full, count_img_edge_continuous = _search_image_edge(
            rgb_mod, candidate_original, candidate_continuous
        )

        if verbose:
            for candidate in (candidate_original, candidate_continuous):
                click.echo(
                    "Candidate value: %s "
                    "Candidate count: %s "
                    "Continuous count: %s"
                    % (
                        str(candidate),
                        str(count_img_edge_full),
                        str(count_img_edge_continuous),
                    )
                )

        output = _evaluate_count(
            count_img_edge_full, count_img_edge_continuous, verbose
        )

        return output

    else:
        raise ValueError("Invalid candidate list {!r}".format(candidate_list))


def determine_nodata(src_path, user_nodata, discovery, debug, verbose):
    """Worker function for determining nodata

    Parameters
    ----------
    src_path: string
    user_nodata: string/integer
                 User supplies the nodata value,
                 input a single value or a string of list
    discovery: Boolean
               determines nodata if alpha channel does not exist
               or internal ndv does not exist
    debug: Boolean
           Enables matplotlib & printing of figures
    verbose: Boolean
             Prints extra information, like competing candidate values


    Returns
    -------
    nodata value: string
                  string(int) or stringified array of values of
                  len == the number of bands.
                  For example, string([int(ndv), int(ndv), int(ndv)])
    """

    with rasterio.open(src_path, "r") as src:
        count = src.count
        data = np.rollaxis(src.read(), 0, 3)

    if user_nodata:
        return user_nodata

    if count == 4:
        return "alpha"
    else:
        nodata = src.nodata
        if nodata is None:
            if discovery:
                candidates = discover_ndv(data, debug, verbose)
                if len(candidates) != 3:
                    return ""
                else:
                    return "[{}, {}, {}]".format(*candidates)
            else:
                return ""
        else:
            return "%s" % (str(int(nodata)))



"""Utility functions"""

from __future__ import division
import json
import math
import re

import numpy as np
from scipy.stats import mode


def _parse_single(n):
    """Returns a single value nodata of type float

    Parameters
    ----------
    n: integer or str(integer)

    Returns
    -------
    float(n)

    """
    try:
        return float(n)
    except ValueError:
        raise ValueError("{0} is not a valid nodata value".format(n))


def _parse_ndv(ndv, bands):
    """Returns a list of nodata values of type float

    Parameters
    ----------
    ndv: string, str(list of nodata values)
    bands: integer, band count

    Returns
    -------
    list: list of floats, length = band count

    """

    if re.match(r"\[[0-9\.\,\s]+\]", ndv):
        ndvals = [_parse_single(n) for n in json.loads(ndv)]
        if len(ndvals) != bands:
            raise ValueError(
                "{0} parsed to ndv of {1} does "
                "not match band count of {2}".format(ndv, json.dumps(ndvals), bands)
            )
        else:
            return ndvals
    else:
        return [_parse_single(ndv) for i in range(bands)]


def _convert_rgb(rgb_orig):
    # Sample to ~200 in smaller dimension if > 200 for performance
    if rgb_orig[:, :, 0].shape[0] < rgb_orig[:, :, 0].shape[1]:
        min_dimension = 0
    else:
        min_dimension = 1

    if rgb_orig[:, :, 0].shape[min_dimension] < 200:
        mod = 1
    else:
        mod = int(math.ceil(rgb_orig[:, :, 0].shape[min_dimension] / 200))

    rgb_mod = rgb_orig[::mod, ::mod]
    # Flatten image for full img histogram
    rgb_mod_flat = rgb_mod.reshape(
        (rgb_mod.shape[0] * rgb_mod.shape[1], rgb_mod.shape[-1])
    )

    return rgb_mod, rgb_mod_flat


# Squish array to only continuous values, return is in list form
def _find_continuous_rgb(input_array, axis_num):
    diff_array = np.diff(input_array, axis=int(axis_num))
    diff_array = np.insert(diff_array, 0, [99, 99, 99], axis=int(axis_num))
    val_list = (input_array[diff_array == [0, 0, 0]]).tolist()
    return val_list


# Find modal RGB value of continuous values array
# (val_list), takes list, returns [R,G,B]
def _group(lst, n, continuous):
    arr = np.asarray(list(zip(*[lst[i::n] for i in range(n)])))
    mode_vals = mode(arr)
    continuous = [int((mode_vals[0])[0, i]) for i in range(3)]
    return continuous, arr


def _compute_continuous(rgb_mod, loc):
    cont_lst = []
    return _group(_find_continuous_rgb(rgb_mod, loc), 3, cont_lst)


def _search_image_edge(rgb_mod, candidate_original, candidate_continuous):
    # Make array of image edge
    top_row = rgb_mod[0, :, :]
    bottom_row = rgb_mod[-1, :, :]
    first_col = rgb_mod[:, 0, :]
    last_col = rgb_mod[:, -1, :]
    img_edge = np.concatenate((top_row, last_col, bottom_row, first_col), axis=0)

    # Squish image edge down to just continuous values
    edge_mode_continuous, arr = _compute_continuous(rgb_mod, 0)

    # Count nodata value frequency in full image edge & squished image edge
    count_img_edge_full = [
        len(np.transpose(np.where((img_edge == candidate).all(axis=1))))
        for candidate in (candidate_original, candidate_continuous)
    ]

    count_img_edge_continuous = [
        len(np.transpose(np.where((arr == candidate).all(axis=1))))
        for candidate in (candidate_original, candidate_continuous)
    ]

    return count_img_edge_full, count_img_edge_continuous


def _evaluate_count(lst1, lst2, verbose):
    # Q: will these always realiably be ordered as listed
    # above with original first, continuous second?
    if (lst1[0] > lst1[1]) and (lst2[0] > lst2[1]):
        return lst1

    elif (lst1[0] < lst1[1]) and (lst2[0] < lst2[1]):
        return lst2

    else:
        if verbose:
            return "None"
        else:
            return ""


def _debug_mode(rgb_flat, arr, output):
    import matplotlib.pyplot as plt

    plt.hist(rgb_flat, bins=range(256))
    # histogram of continuous values only
    plt.hist(arr, bins=range(256))
    plt.savefig(output, bbox_inches="tight")
    plt.close()



"""Alpha masking."""

import numpy as np


def mask_exact(img, ndv):
    """Exact nodata masking based on ndv

    Parameters
    -----------
    img: ndarray
        (depth x rows x cols) array
    ndv: (list|tuple|ndarray)
        list of notdata values where len == img depth

    Returns
    --------
    alpha: ndarray
        ndarray mask of shape (rows, cols) where
        opaque == 0 and transparent == max of dtype
    """
    assert len(ndv) == img.shape[0], "ndv length must equal num bands"
    alpha = np.any(np.transpose(img, [1, 2, 0]) != ndv, axis=2)
    alpha_rescale = alpha.astype(img.dtype) * np.iinfo(img.dtype).max
    return alpha_rescale



"""Setup script."""

import sys

from setuptools import setup, find_packages


with open("rio_alpha/__init__.py") as f:
    for line in f:
        if line.find("__version__") >= 0:
            version = line.split("=")[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            break

long_description = """Rasterio plugin for handling nodata values

See https://github.com/mapbox/rio-alpha for docs.
"""

requirements = ["click>=6", "rasterio~=1.0", "rio-mucho>=1.0rc1", "scipy>=0.18"]

if sys.version_info < (3, 3):
    requirements += ["mock>=2"]

setup(
    name="rio-alpha",
    version=version,
    description=u"A replacement for pxm-alpha",
    long_description=long_description,
    classifiers=[],
    keywords="",
    author=u"Virginia Ng",
    author_email="virginia@mapbox.com",
    url="https://github.com/mapbox/rio-alpha",
    license="BSD",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    include_package_date=True,
    zip_safe=False,
    install_requires=requirements,
    extras_require={
        "plot": ["matplotlib"],
        "test": ["codecov", "hypothesis", "pytest", "pytest-cov"],
    },
    entry_points="""
    [rasterio.rio_plugins]
    alpha=rio_alpha.scripts.cli:alpha
    islossy=rio_alpha.scripts.cli:islossy
    findnodata=rio_alpha.scripts.cli:findnodata
    """,
)



"""Evaluate datasets for lossy compression affected nodata masks"""

from rasterio.features import shapes

from rio_alpha.alpha_mask import mask_exact


def count_ndv_regions(img, ndv):
    """Discover unique labels to count ndv regions.

    Parameters
    ----------
    ndv: list
        a list of floats whose length = band count
    img: ndarray
        (depth x rows x cols) array

    Returns
    -------
    int
        The number of connected regions
    """
    img = mask_exact(img, ndv)
    return len(list(shapes(img, mask=(img != 255))))



"""rio-alpha command."""

import logging
import click

import rasterio as rio
from rio_alpha.utils import _parse_ndv
from rio_alpha.islossy import count_ndv_regions
from rio_alpha.findnodata import determine_nodata
from rio_alpha.alpha import add_alpha
from rasterio.rio.options import creation_options

logger = logging.getLogger("rio_alpha")


@click.command("islossy")
@click.argument("input", nargs=1, type=click.Path(exists=True))
@click.option(
    "--ndv",
    default="[0, 0, 0]",
    help="Expects a string containing a single integer value "
    "(e.g. '255') or "
    "a string representation of a list containing "
    "per-band nodata values (e.g. '[255, 255, 255]').",
)
def islossy(input, ndv):
    """
    Determine if there are >= 10 nodata regions in an image
    If true, returns the string `--lossy lossy`.
    """
    with rio.open(input, "r") as src:
        img = src.read()

    ndv = _parse_ndv(ndv, 3)

    if count_ndv_regions(img, ndv) >= 10:
        click.echo("True")
    else:
        click.echo("False")


@click.command("findnodata")
@click.argument("src_path", type=click.Path(exists=True))
@click.option(
    "--user_nodata",
    "-u",
    default=None,
    help="User supplies the nodata value, "
    "input a string containing a single integer value "
    "(e.g. '255') or "
    "a string representation of a list containing "
    "per-band nodata values (e.g. '[255, 255, 255]').",
)
@click.option(
    "--discovery",
    is_flag=True,
    default=False,
    help="Determines nodata if alpha channel"
    "does not exist or internal ndv does not exist",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="Enables matplotlib & printing of figures",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Prints extra information, " "like competing candidate values",
)
def findnodata(src_path, user_nodata, discovery, debug, verbose):
    """Print a dataset's nodata value."""
    ndv = determine_nodata(src_path, user_nodata, discovery, debug, verbose)
    click.echo("%s" % ndv)


@click.command("alpha")
@click.argument("src_path", type=click.Path(exists=True))
@click.argument("dst_path", type=click.Path(exists=False))
@click.option(
    "--ndv",
    default=None,
    help="Expects a string containing a single integer value "
    "(e.g. '255') or "
    "a string representation of a list containing "
    "per-band nodata values (e.g. '[255, 255, 255]').",
)
@click.option("--workers", "-j", type=int, default=1)
@click.pass_context
@creation_options
def alpha(ctx, src_path, dst_path, ndv, creation_options, workers):
    """Adds/replaced an alpha band to your RGB or RGBA image

    If you don't supply ndv, the alpha mask will be infered.
    """
    with rio.open(src_path) as src:
        band_count = src.count

    if ndv:
        ndv = _parse_ndv(ndv, band_count)

    add_alpha(src_path, dst_path, ndv, creation_options, workers)



"""Alpha masking worker and concurrent processor."""

import numpy as np
import rasterio
import riomucho
from rio_alpha.alpha_mask import mask_exact


def alpha_worker(open_file, window, ij, g_args):
    """rio mucho worker for alpha. It reads input
    files and perform alpha calculations on each window.

    Parameters
    ------------
    open_files: list of rasterio open files
    window: Window object
            A window is a view onto a rectangular subset of a
            raster dataset.
    g_args: dictionary

    Returns
    ---------
    rgba: ndarray
          ndarray with original RGB bands of shape (3, rows, cols)
          and a mask of shape (rows, cols) where
          opaque == 0 and transparent == max of dtype
    """
    src = open_file[0]

    arr = src.read(window=window)

    # Determine Alpha Band
    if g_args["ndv"]:
        # User-supplied nodata value
        alpha = mask_exact(arr, g_args["ndv"])
    else:
        # Let rasterio decide
        alpha = src.dataset_mask(window=window)

    # Replace or Add alpha band to input data
    if arr.shape[0] == 4:
        # replace band 4 with new alpha mask
        # (likely the same but let's not make that assumption)
        rgba = arr.copy()
        rgba[3] = alpha
    elif arr.shape[0] == 3:
        # stack the alpha mask to add band 4
        rgba = np.append(arr, alpha[np.newaxis, :, :], axis=0)
    else:
        raise ValueError("Array must have 3 or 4 bands (RGB or RGBA)")

    return rgba


_alpha_worker = alpha_worker


def add_alpha(src_path, dst_path, ndv, creation_options, processes):
    """
    Parameters
    ------------
    src_paths: list of strings
    dst_path: string
    ndv: list
         a list of floats where the
         length of the list = band count
    creation_options: dict
    processes: integer


    Returns
    ---------
    None
        Output is written to dst_path
    """

    with rasterio.open(src_path) as src:
        dst_profile = src.profile

    dst_profile.update(**creation_options)

    dst_profile.pop("photometric", None)

    dst_profile.update(count=4, nodata=None)

    global_args = {"src_nodata": 0, "dst_dtype": dst_profile["dtype"], "ndv": ndv}

    with riomucho.RioMucho(
        [src_path],
        dst_path,
        _alpha_worker,
        options=dst_profile,
        global_args=global_args,
        mode="manual_read",
    ) as rm:

        rm.run(processes)
